/*
 *  Rewrite of AIX metrics using libperfstat API
 *
 *  Libperfstat can deal with a 32-bit and a 64-bit Kernel and does not require root authority.
 *
 *  This new version incorporates POWER5 specific extensions
 *  (SPLPAR, SMT, CPU Entitlement, Capped/Uncapped etc.)
 *
 *  The code has been tested with AIX 5.1, AIX 5.2 and AIX 5.3 on different systems.
 *
 *  Written by Michael Perzl (michael@perzl.org)
 *         and Nigel Griffiths (nigelargriffiths@hotmail.com)
 *  May 22, 2006
 *
 *  Last updated: Jun 19, 2007
 *
 *  First version of rewritten AIX metrics using the libperfstat API
 *  by Andreas Schoenfeld, TU Darmstadt, Germany (4/2005) 
 *  E-Mail: Schoenfeld@hrz.tu-darmstadt.de
 *
 *  Its based on the 
 *  First stab at support for metrics in AIX
 *  by Preston Smith <psmith@physics.purdue.edu>
 *  Wed Feb 27 14:55:33 EST 2002
 *
 *  AIX V5 support, bugfixes added by Davide Tacchella <tack@cscs.ch>
 *  May 10, 2002
 *
 *  you may still find some code (like "int bos_level(..)"  ) and the basic structure of this version. 
 *
 */

#include "interface.h"
#include <stdlib.h>
#include <ctype.h>
#include <utmp.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <procinfo.h>
#include <strings.h>
#include <signal.h>
#include <odmi.h>
#include <cf.h>
#include <sys/utsname.h>

#if !defined(_AIX43)
#include <sys/dr.h>
#endif
#include <sys/systemcfg.h>

#include <libperfstat.h>

#include "libmetrics.h"

#ifndef FIRST_NETINTERFACE
#define FIRST_NETINTERFACE ""
#endif

#if defined(_AIX43)
#ifndef SBITS
/*
 * For multiplication of fractions that are stored as integers, including
 * p_pctcpu.  Not allowed to do floating point arithmetic in the kernel.
 */
#define SBITS   16
#endif
#endif

#define CPUINFO_TIMEOUT		5.0
#define NETWORKINFO_TIMEOUT	5.0

#define MEM_KB_PER_PAGE (4096/1024)



struct product {
        __long32_t _id;         /* offset: 0x00 (   0) */
        __long32_t _reserved;   /* offset: 0x04 (   4) */
        __long32_t _scratch;    /* offset: 0x08 (   8) */
        char lpp_name[145];     /* offset: 0x0c (  12) */
        char comp_id[20];       /* offset: 0x9d ( 157) */
        short update;           /* offset: 0xb2 ( 178) */
        __long32_t cp_flag;     /* offset: 0xb4 ( 180) */
        char fesn[10];          /* offset: 0xb8 ( 184) */
        char *name;             /* offset: 0xc4 ( 196) */
        short state;            /* offset: 0xc8 ( 200) */
        short ver;              /* offset: 0xca ( 202) */
        short rel;              /* offset: 0xcc ( 204) */
        short mod;              /* offset: 0xce ( 206) */
        short fix;              /* offset: 0xd0 ( 208) */
        char ptf[10];           /* offset: 0xd2 ( 210) */
        short media;            /* offset: 0xdc ( 220) */
        char sceded_by[10];     /* offset: 0xde ( 222) */
        char *fixinfo;          /* offset: 0xe8 ( 232) */
        char *prereq;           /* offset: 0xec ( 236) */
        char *description;      /* offset: 0xf0 ( 240) */
        char *supersedes;       /* offset: 0xf4 ( 244) */
};



static int aixVersion, aixRelease, aixLevel, aixFix;

static int isVIOserver;

static time_t boottime;

struct cpuinfo_stats {
                        double user;
                        double sys;
                        double idle;
                        double wait;
                     } current_cpuinfo;

struct network_stats {
                        double ibytes;
                        double obytes;
                        double ipackets;
                        double opackets;
                     } current_network_stats;



g_val_t
cpu_speed_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.uint32 = 0;
   else
      val.uint32 = c.processorHZ / 1000000;

   return( val );
}


static time_t
boottime_func_CALLED_ONCE( void )
{
   time_t boottime;
   struct utmp buf;
   FILE *utmp;


   utmp = fopen( UTMP_FILE, "r" );

   if (utmp == NULL)
   {
      /* Can't open utmp, use current time as boottime */
      boottime = time( NULL );
   }
   else
   {
      while (fread( (char *) &buf, sizeof( buf ), 1, utmp ) == 1)
      {
         if (buf.ut_type == BOOT_TIME)
         {
            boottime = buf.ut_time;
            break;
        }
      }

      fclose( utmp );
   }

   return( boottime );
}



g_val_t
boottime_func( void )
{
   g_val_t val;


   val.uint32 = boottime;

   return( val );
}



g_val_t
sys_clock_func( void )
{
   g_val_t val;

   val.uint32 = time( NULL );

   return( val );
}



g_val_t
machine_type_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      strcpy( val.str, "Unknown type" );
   else
     strncpy( val.str, c.description, MAX_G_STRING_SIZE );

   return( val );
}



g_val_t
os_name_func( void )
{
   g_val_t val;
   struct utsname uts;


   if (isVIOserver)
      strcpy( val.str, "Virtual I/O Server" );
   else
   {
      uname ( &uts );
      strncpy( val.str, uts.sysname, MAX_G_STRING_SIZE );
   }

   return( val );
}        



g_val_t
os_release_func( void )
{
   g_val_t val;
   char oslevel[MAX_G_STRING_SIZE];


   sprintf( oslevel, "%d.%d.%d.%d", aixVersion, aixRelease, aixLevel, aixFix );
   strncpy( val.str, oslevel, MAX_G_STRING_SIZE );

   return( val );
}        



/*
 * AIX  defines:
 * CPU_USER, CPU_SYS, CPU_IDLE, CPU_WAIT
 * so no metrics for cpu_nice, or cpu_aidle
 */



void
update_cpuinfo( void )
{
   perfstat_cpu_total_t c;
   static u_longlong_t last_user_ticks = 0L;
   static u_longlong_t last_sys_ticks = 0L;
   static u_longlong_t last_idle_ticks = 0L;
   static u_longlong_t last_wait_ticks = 0L;
   static u_longlong_t last_total_ticks = 0L;
   u_longlong_t user_ticks, sys_ticks, idle_ticks, wait_ticks,
                total_ticks, diff;
   static double last_time = 0.0;
   double now;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);


/* Do nothing if we are called too soon after the last call */

   if (now - last_time < CPUINFO_TIMEOUT) return;

   last_time = now;


/* obtain CPU parameters */

   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) != -1)
   {
      user_ticks = c.user;
      sys_ticks = c.sys;
      idle_ticks = c.idle;
      wait_ticks = c.wait;
      total_ticks = c.user + c.sys + c.idle + c.wait;

      diff = total_ticks - last_total_ticks;

      if ( diff )
      {
         current_cpuinfo.user = 100.0 * (user_ticks - last_user_ticks) / diff;
         current_cpuinfo.sys = 100.0 * (sys_ticks - last_sys_ticks) / diff;
         current_cpuinfo.idle = 100.0 * (idle_ticks - last_idle_ticks) / diff;
         current_cpuinfo.wait = 100.0 * (wait_ticks - last_wait_ticks) / diff;
      }

      last_user_ticks = user_ticks;
      last_sys_ticks = sys_ticks;
      last_idle_ticks = idle_ticks;
      last_wait_ticks = wait_ticks;
      last_total_ticks = total_ticks;
   }
}



g_val_t
cpu_user_func( void )
{
   g_val_t val;


   update_cpuinfo();

   val.f = current_cpuinfo.user;

   return( val );
}



g_val_t
cpu_system_func( void )
{
   g_val_t val;


   update_cpuinfo();

   val.f = current_cpuinfo.sys;

   return( val );
}



g_val_t
cpu_idle_func( void )
{
   g_val_t val;


   update_cpuinfo();

   val.f = current_cpuinfo.idle;

   return( val );
}



g_val_t
cpu_wio_func( void )
{
   g_val_t val;


   update_cpuinfo();

   val.f = current_cpuinfo.wait;

   return( val );
}



/*
 * AIX does not have metrics for CPU aidle
 */

g_val_t 
cpu_aidle_func( void )
{
   g_val_t val;


   val.f = 0.0;

   return( val );
}



/*
 * AIX does not have metrics for CPU nice 
 */

g_val_t
cpu_nice_func( void )
{
   g_val_t val;


   val.f = 0.0;

   return( val );
}



/*
 * FIXME - Find number of interrupts (?) 
 */
g_val_t 
cpu_intr_func( void )
{
   g_val_t val;


   val.f = 0.0;

   return( val );
}



/* 
 * FIXME - find number of system interrupts (?) 
 */

g_val_t 
cpu_sintr_func( void )
{
   g_val_t val;


   val.f = 0.0;
   
   return( val );
}



void
update_ifdata( void )
{
   static int init_done = 0;
   perfstat_id_t name;
   perfstat_netinterface_t *nif_buf, *p;
   static u_longlong_t last_bytes_in = 0L;
   static u_longlong_t last_bytes_out = 0L;
   static u_longlong_t last_pkts_in = 0L;
   static u_longlong_t last_pkts_out = 0L;
   u_longlong_t bytes_in, bytes_out, pkts_in, pkts_out;
   longlong_t d;
   int i, nr_netif;
   static double last_time = 0.0;
   double now, delta_t;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);

   delta_t = now - last_time;


/* Do nothing if we are called too soon after the last call */

   if (init_done && (delta_t < NETWORKINFO_TIMEOUT)) return;


/* obtain the network parameters */

   bytes_in = bytes_out = pkts_in = pkts_out = 0L;

   nr_netif = perfstat_netinterface( NULL, NULL, sizeof( perfstat_netinterface_t ), 0 );

   if (nr_netif == -1)
      return;
   else
   {
/* We can allocate enough memory for the buffer */
      nif_buf = (perfstat_netinterface_t *)
                    malloc( nr_netif * sizeof( perfstat_netinterface_t));

      if (nif_buf == NULL)
/* couldn't allocate enough memory */
         return;
      else
      { 
/* In order to get all of the network interfaces statistics,
 * we will first ask for the network interface which name is "",
 * that can be considered as an alias for the name of the first
 * network interface.
 */
         strcpy( name.name, FIRST_NETINTERFACE );

         if (perfstat_netinterface( &name,
                                    nif_buf,
                                    sizeof( perfstat_netinterface_t ),
                                    nr_netif ) == -1)
         {
            free( nif_buf );
            return;
         }
         else
         {
/* Ok, we could get the statistics, we can now use it. */
/* Ignore all loopback devices */
            for (i = 0, p = nif_buf;  i < nr_netif;  i++, p++ )
               if (p->type != IFT_LOOP)
               {
                  bytes_in  += p->ibytes;
                  bytes_out += p->obytes;
                  pkts_in   += p->ipackets;
                  pkts_out  += p->opackets;
               }
         }

/* free allocated memory again */
         free( nif_buf );
      }
   }


/* calculate the network parameters */

   if (init_done)
   {
/* get the number of bytes transferred in, check for integer overrun */
      d = bytes_in - last_bytes_in;
      if (d < 0) d += ULONG_MAX;
      if ( delta_t )
         current_network_stats.ibytes = (double) d / delta_t;
      else
         current_network_stats.ibytes = 0.0;

      last_bytes_in = bytes_in;

/* get the number of bytes transferred out, check for integer overrun */
      d = bytes_out - last_bytes_out;
      if (d < 0) d += ULONG_MAX;

      if ( delta_t )
         current_network_stats.obytes = (double) d / delta_t;
      else
         current_network_stats.obytes = 0.0;

      last_bytes_out = bytes_out;

/* get the number of packets transferred in, check for integer overrun */
      d = pkts_in - last_pkts_in;
      if (d < 0) d += ULONG_MAX;

      if ( delta_t )
         current_network_stats.ipackets = (double) d / delta_t;
      else
         current_network_stats.ipackets = 0.0;

      last_pkts_in = pkts_in;

/* get the number of packets transferred out, check for integer overrun */
      d = pkts_out - last_pkts_out;
      if (d < 0) d += ULONG_MAX;

      if ( delta_t )
         current_network_stats.opackets = (double) d / delta_t;
      else
         current_network_stats.opackets = 0.0;

      last_pkts_out = pkts_out;
   }
   else
   {
      init_done = 1;

      current_network_stats.ibytes   = 0.0;
      current_network_stats.obytes   = 0.0;
      current_network_stats.ipackets = 0.0;
      current_network_stats.opackets = 0.0;

      last_bytes_in  = bytes_in;
      last_bytes_out = bytes_out;
      last_pkts_in   = pkts_in;
      last_pkts_out  = pkts_out;
   }

   last_time = now;
}



g_val_t
bytes_in_func( void )
{
   g_val_t val;


   update_ifdata();

   val.f = current_network_stats.ibytes;

   return( val );
}



g_val_t
bytes_out_func( void )
{
   g_val_t val;


   update_ifdata();

   val.f = current_network_stats.obytes;

   return( val );
}



g_val_t
pkts_in_func( void )
{
   g_val_t val;


   update_ifdata();

   val.f = current_network_stats.ipackets;

   return( val );
}



g_val_t
pkts_out_func( void )
{
   g_val_t val;


   update_ifdata();

   val.f = current_network_stats.opackets;

   return( val );
}



g_val_t 
disk_free_func( void )
{
   g_val_t val;
   perfstat_disk_total_t d;


   if (perfstat_disk_total( NULL, &d, sizeof( perfstat_disk_total_t ), 1 ) == -1)
      val.d = 0.0;
   else
      val.d = (double) d.free;

   return( val );
}



g_val_t 
disk_total_func( void )
{
   g_val_t val;
   perfstat_disk_total_t d;


   if (perfstat_disk_total( NULL, &d, sizeof( perfstat_disk_total_t ), 1 ) == -1)
      val.d = 0.0;
   else
      val.d = (double) d.size;

   return( val);
}


/*
 * FIXME: Find disk partition which is the most full one
 */

g_val_t
part_max_used_func( void )
{
   g_val_t val;


   val.f = 0.0;

   return( val );
}



g_val_t
load_one_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
      val.f = (double) c.loadavg[0] / (double) (1 << SBITS);

   return( val );
}



g_val_t
load_five_func( void )
{  
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
      val.f = (double) c.loadavg[1] / (double) (1 << SBITS);

   return( val );
}



g_val_t
load_fifteen_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
      val.f = (double) c.loadavg[2] / (double) (1 << SBITS);

   return( val );
}



g_val_t
cpu_num_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.uint16 = 0;
   else
      val.uint16 = c.ncpus;

   return( val );
}



g_val_t
proc_total_func ( void )
{
  pid_t IndexPointer;
  int n;
  g_val_t val;


  IndexPointer = (pid_t) 0;
  n = getprocs( NULL, 0, NULL, 0, &IndexPointer, 10000000 );

  if (n == -1) n = 0;

  val.uint32 = n;
 
  return( val );
}



g_val_t
proc_run_func( void )
{
   g_val_t val;
   perfstat_cpu_total_t c;
   static u_longlong_t last_runque = 0;
   static double last_time = 0.0;
   double now, delta_t, runnable;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);

   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      runnable = 0.0;
   else
   {
      delta_t = now - last_time;

      if (delta_t != 0.0)
         runnable = (double) (c.runque - last_runque + 0.5) / delta_t;
      else
         runnable = 0.0;   

      last_runque = c.runque;
   }

   last_time = now;

   val.uint32 = (unsigned int) runnable;
 
   return( val );
}



g_val_t
mem_total_func( void )
{
   g_val_t val;
   perfstat_memory_total_t m;


   if (perfstat_memory_total( NULL, &m, sizeof( perfstat_memory_total_t ), 1 ) == -1)
      val.uint32 = 0;
   else
      val.uint32 = m.real_total * MEM_KB_PER_PAGE;

   return( val );
}



g_val_t
mem_free_func( void )
{
   g_val_t val;
   perfstat_memory_total_t m;


   if (perfstat_memory_total( NULL, &m, sizeof( perfstat_memory_total_t ), 1 ) == -1)
      val.uint32 = 0;
   else
      val.uint32 = m.real_free * MEM_KB_PER_PAGE; 

   return( val );
}



g_val_t
mem_shared_func( void )
{
   g_val_t val;

   val.uint32 = 0;

   return( val );
}



g_val_t
mem_buffers_func( void )
{
   g_val_t val;

   val.uint32 = 0;

   return( val );
}



g_val_t
mem_cached_func( void )
{
   g_val_t val;
   perfstat_memory_total_t m;


   if (perfstat_memory_total( NULL, &m, sizeof( perfstat_memory_total_t ), 1 ) == -1)
      val.uint32 = 0;
   else
      val.uint32 = m.numperm * MEM_KB_PER_PAGE;

   return( val );
}



g_val_t
swap_total_func( void )
{
   g_val_t val;
   perfstat_memory_total_t m;


   if (perfstat_memory_total( NULL, &m, sizeof( perfstat_memory_total_t ), 1 ) == -1)
      val.uint32 = 0;
   else
      val.uint32 = m.pgsp_total * MEM_KB_PER_PAGE;

   return( val );
}



g_val_t
swap_free_func( void )
{
   g_val_t val;
   perfstat_memory_total_t m;


   if (perfstat_memory_total( NULL, &m, sizeof( perfstat_memory_total_t ), 1) == -1)
      val.uint32 = 0;
   else
      val.uint32 = m.pgsp_free * MEM_KB_PER_PAGE;

   return( val );
}



/* We want to find the minimum MTU (Max packet size) over all UP interfaces. */
g_val_t
mtu_func( void )
{
   g_val_t val;


   val.uint32 = get_min_mtu();

/* A value of 0 means there are no UP interfaces. Shouldn't happen. */
   return( val );
}



/*  int bos_level(int *aix_version, int *aix_release, int *aix_level, int *aix_fix)
 *  is copied from 
 *
 *  First stab at support for metrics in AIX
 *  by Preston Smith <psmith@physics.purdue.edu>
 *  Wed Feb 27 14:55:33 EST 2002
 *
 *  AIX V5 support, bugfixes added by Davide Tacchella <tack@cscs.ch>
 *  May 10, 2002
 *
 *  AIX 5L adaptation done by Michael Perzl (michael@perzl.org)
 *  March 30, 2006
 */

static int
get_OS_Level_Information( int *AIX_Version,
                          int *AIX_Release,
                          int *AIX_Level,
                          int *AIX_Fix )
{
   struct Class *product_CLASS, *productClass;
   struct product productObj;
   int maxVer, maxRel, maxMod, maxFix;
   int rc, getit, found = 0;
   char *path;


/* Prepare the ODM for use with an application program */

   if (odm_initialize() == -1)
      return( E_ODMINIT ); 


/* Make sure we take the right database */

   if ((path = odm_set_path( "/usr/lib/objrepos" )) == (char *) -1)
      return( odmerrno );

 
/* Mount the 'product' class */

   if ((product_CLASS = odm_mount_class( "product" )) == (CLASS_SYMBOL) -1)
      return( odmerrno );


/* Open the ODM 'product' class */

   if ((int) (productClass = odm_open_class( product_CLASS )) == -1)
      return( E_ODMOPEN );

/* Loop trough all entries for the product name. */
/* We look for the following filesets:           */
/*   - bos.up and bos.up                         */
/*   - bos.up64 and bos.mp64                     */

   maxVer = maxRel = maxMod = maxFix = -1;

   getit = ODM_FIRST;
   while ((rc = (int) odm_get_obj( productClass, "name like bos.?p",
                                   &productObj, getit)) != 0)
   {
      getit = ODM_NEXT;
      if (rc == -1)
      {
         /* no (more) ODM objects found... */
         break;
      }
      else
      {
         if (productObj.ver > maxVer) maxVer = productObj.ver;
         if (productObj.rel > maxRel) maxRel = productObj.rel;
         if (productObj.mod > maxMod) maxMod = productObj.mod;
         if (productObj.fix > maxFix) maxFix = productObj.fix;
         found++;
      }
   }

   getit = ODM_FIRST;
   while ((rc = (int) odm_get_obj( productClass, "name like bos.?p64",
                                   &productObj, getit)) != 0)
   {
      getit = ODM_NEXT;
      if (rc == -1)
      {
         /* no (more) ODM objects found... */
         break;
      }
      else
      {
         if (productObj.ver > maxVer) maxVer = productObj.ver;
         if (productObj.rel > maxRel) maxRel = productObj.rel;
         if (productObj.mod > maxMod) maxMod = productObj.mod;
         if (productObj.fix > maxFix) maxFix = productObj.fix;
         found++;
      }
   }

   if (! found)
      maxVer = maxRel = maxMod = maxFix = 0;

   if (found)
   {
      *AIX_Version = maxVer;
      *AIX_Release = maxRel;
      *AIX_Level   = maxMod;
      *AIX_Fix     = maxFix;
   }


/* Close the ODM object class */

   odm_close_class( productClass );


/* Terminate the ODM session */

   odm_terminate();


   free( path );

   return( found ? 0 : -1 );
}



g_val_t
capped_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      strcpy( val.str, "libperfstat returned an error" );
   else
      if ( __LPAR() && p.type.b.shared_enabled )
         strcpy ( val.str, p.type.b.capped ? "yes" : "no" );
      else
         strcpy( val.str, "No SPLPAR-capable system" );
#else
   strcpy( val.str, "No SPLPAR-capable system" );
#endif

   return( val );
}



g_val_t
cpu_entitlement_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
      if ( p.type.b.shared_enabled )
      {
         val.f = p.entitled_proc_capacity / 100.0;
      }
      else /* dedicated LPAR/standalone system so fake entitled as number of online CPUs */
         val.f = p.online_cpus;
#else
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.f = 0.;
   else
      val.f = c.ncpus;
#endif

   return( val );
}



g_val_t
cpu_in_lpar_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = p.online_cpus;
#else
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = c.ncpus_cfg;
#endif

   return( val );
}



g_val_t
cpu_in_machine_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = p.online_phys_cpus_sys;
#else
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = c.ncpus;
#endif

   return( val );
}



g_val_t
cpu_in_pool_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = p.phys_cpus_pool;
#else
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.int32 = -1;
   else
      val.int32 = c.ncpus;
#endif

   return( val );
}



g_val_t
cpu_pool_idle_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;
   lpar_info_format2_t f2;
   static uint64_t saved_pool_idle_time = 0;
   static double last_time = 0.0;
   double now, delta_t;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);
 
   lpar_get_info( LPAR_INFO_FORMAT2, &f2, sizeof( lpar_info_format2_t ) );

   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
   {
      delta_t = now - last_time;

      if ( p.type.b.shared_enabled )
      {
         if ( (delta_t != 0.0) && (f2.lpar_flags & LPAR_INFO2_AUTH_PIC) )
            val.f = (double) (f2.pool_idle_time - saved_pool_idle_time) /
                    (double) delta_t / 1000.0 / 1000.0 / 1000.0;
         else
            val.f = 0.0;

         saved_pool_idle_time = f2.pool_idle_time;
      }
      else
         val.f = 0.0;
   }

   last_time = now;
#else
   val.f = 0.0;
#endif

   return( val );
}



g_val_t
cpu_used_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;
   lpar_info_format2_t f2;
   static uint64_t saved_dispatch_time = 0;
   static double last_time = 0.0;
   double now, delta_t;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);

   if (lpar_get_info( LPAR_INFO_FORMAT2, &f2, sizeof( lpar_info_format2_t ) ) == -1)
      val.f = 0.0;
   else
      if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1 ) == -1)
         val.f = 0.0;
      else
      {
         delta_t = now - last_time;

         if ( p.type.b.shared_enabled )
         {
            if ( delta_t != 0.0 )
               val.f = (double) (f2.tot_dispatch_time - saved_dispatch_time) /
                       (double) delta_t / 1000.0 / 1000.0 / 1000.0;
            else
               val.f = 0.0;

            saved_dispatch_time = f2.tot_dispatch_time;
         }
         else /* dedicated LPAR/standalone system so fake entitled as number of online CPUs */
            val.f = p.online_cpus;
      }

   last_time = now;
#else
   perfstat_cpu_total_t c;


   if (perfstat_cpu_total( NULL, &c, sizeof( perfstat_cpu_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
      val.f = c.ncpus_cfg;
#endif

   return( val );
}



g_val_t
disk_read_func( void )
{
   g_val_t val;
   static perfstat_disk_total_t d1, d2;
   static double last_time = 0.0;
   double now, delta_t;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);
   if (perfstat_disk_total( NULL, &d2, sizeof( perfstat_disk_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
   {
      delta_t = now - last_time;

      if ( delta_t != 0.0 )
/* the result is returned in number of 512 byte blocks */
         val.f = (double) (d2.rblks - d1.rblks) / 2.0 / (double) delta_t;
      else
         val.f = 0.0;

      d1 = d2;
   }

   last_time = now;

   return( val );
}



g_val_t
disk_write_func( void )
{
   g_val_t val;
   static perfstat_disk_total_t d1, d2;
   static double last_time = 0.0;
   double now, delta_t;
   struct timeval timeValue;
   struct timezone timeZone;


   gettimeofday( &timeValue, &timeZone );

   now = (double) (timeValue.tv_sec - boottime) + (timeValue.tv_usec / 1000000.0);

   if (perfstat_disk_total( NULL, &d2, sizeof( perfstat_disk_total_t ), 1 ) == -1)
      val.f = 0.0;
   else
   {
      delta_t = now - last_time;

      if ( delta_t != 0.0 )
/* the result is returned in number of 512 byte blocks */
         val.f = (double) (d2.wblks - d1.wblks) / 2.0 / (double) delta_t;
      else
         val.f = 0.0;

      d1 = d2;
   }

   last_time = now;

   return( val );
}



g_val_t
kernel64bit_func( void )
{
   g_val_t val;

   strcpy ( val.str, __KERNEL_64() ? "yes" : "no" );

   return( val );
}



g_val_t
lpar_func( void )
{
   g_val_t val;

#if defined(_AIX43)
   strcpy ( val.str, "no" );
#else
   strcpy ( val.str, __LPAR() ? "yes" : "no" );
#endif

   return( val );
}



g_val_t
lpar_name_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if ( __LPAR() )
   {
      if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1) == -1)
         strcpy( val.str, "libperfstat returned an error" );
      else
         strcpy( val.str, p.name );
   }
   else
      strcpy( val.str, "No LPAR system" );
#else
   FILE *f;
   char  buf[64], *p;

#if defined(_AIX43)
   if ( 0 )
#else
   if ( __LPAR() )
#endif
   {
      f = popen( "/usr/bin/uname -L", "r" );

      if (f == NULL)
         strcpy( val.str, "popen of cmd uname -L failed" );
      else
         if ( fgets( buf, 64, f ) != NULL)
         {
            p = &buf[0];
            while (*p != ' ') p++;
            p++;

            strncpy( val.str, p, MAX_G_STRING_SIZE );
            val.str[MAX_G_STRING_SIZE - 1] = '\0';
            val.str[strlen( val.str ) - 1] = '\0';  /* truncate \n */
         }
         else
            strcpy( val.str, "Can't run uname -L" );
   }
   else
      strcpy( val.str, "No LPAR system" );
#endif

   return( val );
}



g_val_t
lpar_num_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t ), 1) == -1)
      val.int32 = -1;
   else
      val.int32 = p.lpar_id;
#else
   FILE *f;
   char s[MAX_G_STRING_SIZE];

#if defined(_AIX43)
   if ( 0 )
#else
   if ( __LPAR() )
#endif
   {
      f = popen( "/usr/bin/uname -L 2>/dev/null", "r" );
      if (f == NULL)
         val.int32 = -1;
      else
         if ( fgets( s, MAX_G_STRING_SIZE, f ) != NULL)
            val.int32 = atoi( s );
         else
            val.int32 = -1;
   }
   else
      val.int32 = -1;
#endif

   return( val );
}



static g_val_t
oslevel_func_CALLED_ONCE( void )
{
   FILE    *f;
   g_val_t  val;


   if (isVIOserver)
   {
      f = popen( "/usr/ios/cli/ioscli ioslevel 2>/dev/null", "r" );

      if (f == NULL)
      {
         strcpy( val.str, "popen of cmd ioscli failed" );
      }
      else
      {
         if ( fgets( val.str, MAX_G_STRING_SIZE, f ) != NULL )
         {
            val.str[MAX_G_STRING_SIZE - 1] = '\0';
            val.str[strlen( val.str ) - 1] = '\0';  /* truncate \n */
         }
         else
            strcpy( val.str, "Can't run AIX cmd ioscli" );

         pclose( f );
      }
   }
   else
   {
      f = popen( "/usr/bin/oslevel -s 2>/dev/null", "r" );

      if (f == NULL)
      {
         strcpy( val.str, "popen of cmd oslevel -s failed" );
      }
      else
      {
         if ( fgets( val.str, MAX_G_STRING_SIZE, f ) != NULL )
         {
            val.str[MAX_G_STRING_SIZE - 1] = '\0';
            val.str[strlen( val.str ) - 1] = '\0';  /* truncate \n */
         }
         else
            strcpy( val.str, "Can't run AIX cmd oslevel" );

         if (! strncmp( val.str, "Usage: oslevel", 14 ))
         {
            pclose( f );

            f = popen( "/usr/bin/oslevel -r 2>/dev/null", "r" );
      
            if ( fgets( val.str, MAX_G_STRING_SIZE, f ) != NULL )
            {
               val.str[MAX_G_STRING_SIZE - 1] = '\0';
               val.str[strlen( val.str ) - 1] = '\0';  /* truncate \n */
            }
            else
               strcpy( val.str, "Can't run AIX cmd oslevel" );
         }

         pclose( f );
      }
   }

   return( val );
}



g_val_t
oslevel_func( void )
{
   static g_val_t val;
   static int firstTime = 1;


   if (firstTime)
   {
      val = oslevel_func_CALLED_ONCE();
      firstTime = 0;
   }

   return( val );
}



static g_val_t
serial_num_func_CALLED_ONCE( void )
{
   FILE    *f;
   g_val_t  val;


   f = popen( "/usr/bin/uname -u 2>/dev/null", "r" );

   if (f == NULL)
   {
      strcpy( val.str, "popen of cmd uname -u failed" );
   }
   else
   {
      if ( fgets( val.str, MAX_G_STRING_SIZE, f ) != NULL )
      {
         val.str[MAX_G_STRING_SIZE - 1] = '\0';
         val.str[strlen( val.str ) - 1] = '\0';  /* truncate \n */
      }
      else
      {
         strcpy( val.str, "Can't run AIX cmd uname -u" );
      }

      pclose( f );
   }


   return( val );
}



g_val_t
serial_num_func( void )
{
   static g_val_t val;
   static int firstTime = 1;


   if (firstTime)
   {
      val = serial_num_func_CALLED_ONCE();
      firstTime = 0;
   }

   return( val );
}



g_val_t
smt_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t), 1 ) == -1)
      strcpy( val.str, "libperfstat returned an error" );
   else
      strcpy( val.str, p.type.b.smt_enabled ? "yes" : "no" );
#else
   strcpy( val.str, "No SMT-capable system" );
#endif

   return( val );
}



g_val_t
splpar_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t), 1) == -1)
      strcpy( val.str, "libperfstat returned an error" );
   else
      strcpy( val.str, p.type.b.shared_enabled ? "yes" : "no" );
#else
   strcpy( val.str, "No SPLPAR-capable system" );
#endif

   return( val );
}



g_val_t
weight_func( void )
{
   g_val_t val;
#if defined(_AIX53)
   perfstat_partition_total_t p;


   if (perfstat_partition_total( NULL, &p, sizeof( perfstat_partition_total_t), 1 ) == -1)
      val.int32 = -1;
   else
      if ( p.type.b.shared_enabled )
         val.int32 = p.var_proc_capacity_weight;
      else
         val.int32 = -1;
#else
   val.int32 = -1;
#endif

   return( val );
}



/*
 * This function is called only once by the gmond.  Use to 
 * initialize data structures, etc or just return SYNAPSE_SUCCESS;
 */
g_val_t
metric_init( void )
{
   g_val_t val;
   FILE *f;


/* get OS level information from ODM */

   (void) get_OS_Level_Information( &aixVersion, &aixRelease,
                                    &aixLevel, &aixFix );


/* find out if we are running on a VIO server */

   f = fopen( "/usr/ios/cli/ioscli", "r" );

   if (f)
   {
      isVIOserver = 1;
      fclose( f );
   }
   else
      isVIOserver = 0;


/* initialize the routines which require a time interval */

   boottime = boottime_func_CALLED_ONCE();
   update_cpuinfo();
   update_ifdata();
   val = proc_run_func();
   val = cpu_pool_idle_func();
   val = cpu_used_func();
   val = disk_read_func();
   val = disk_write_func();


   val.int32 = SYNAPSE_SUCCESS;

   return( val );
}

